import os
import boto3
import openai
from paperqa import Docs
from urllib.parse import urlparse
from typing import Dict, List, Union
from langchain.base_language import BaseLanguageModel
from langchain.chat_models import ChatOpenAI, ChatAnthropic

from pydantic import BaseModel, Field

class KnowledgeBase(Docs, BaseModel):
    llms: Dict[str, BaseLanguageModel] = Field(default_factory=dict)
    documents: List[str] = Field(default_factory=list)
    temperature: float = Field(0.7)

    def __init__(self, documents: List[str], *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.llms = {}
        self.documents = documents
        self.temperature = kwargs.get('temperature', 0.7)

        # Always Initialize both LLMs
        self.llms['gpt4'] = ChatOpenAI(model='gpt-4', temperature=self.temperature)
        self.llms['claude'] = ChatAnthropic(model="claude-2", temperature=self.temperature)
        self.llms['both'] = Docs(llm=self.llms['gpt4'], summary_llm=self.llms['claude'])

        # Process documents for each LLM
        for llm in self.llms.values():
            self.process_documents(llm)

    def process_documents(self, llm: BaseLanguageModel):
        # Add documents to Paper QA
        for path in self.documents:
            parsed_url = urlparse(path)
            scheme = parsed_url.scheme
            ext = os.path.splitext(path)[-1].lower()

            if scheme in ['http', 'https']:
                self.add_url(path)
            elif scheme == 's3':
                s3 = boto3.client('s3')
                bucket = parsed_url.netloc
                key = parsed_url.path.lstrip('/')

                if key.endswith('/'):  # It's a directory
                    try:
                        objects = s3.list_objects(Bucket=bucket, Prefix=key)
                        for obj in objects.get('Contents', []):
                            file_key = obj['Key']
                            local_path = f"/tmp/{os.path.basename(file_key)}"
                            s3.download_file(bucket, file_key, local_path)
                            self.add(local_path)
                    except NoCredentialsError:
                        print("No AWS credentials found.")
                else:  # It's a single file
                    local_path = f"/tmp/{os.path.basename(key)}"
                    s3.download_file(bucket, key, local_path)

                    if ext == '.pdf':
                        self.add(local_path)
                    else:
                        with open(local_path, 'r') as f:
                            text_content = f.read()
                            self.add_file(f, text_content)
            else:
                if ext == '.pdf':
                    self.add(path)
                else:
                    with open(path, 'r') as f:
                        text_content = f.read()
                        self.add_file(f, text_content)


import json
import importlib
from langchain import PromptTemplate

class AspectFactory:
    def __init__(self, llm=None, ddb_util=None, google_util=None, context=None, aspect=None):

        DDBUtilWrapper = importlib.import_module('autopop.ddb_util_wrapper').DDBUtilWrapper
        GoogleUtilWrapper = importlib.import_module('autopop.google_util_wrapper').GoogleUtilWrapper

        self.llm = llm
        self.aspect = aspect
        self.context = context
        self.gu = google_util
        
        # load data from DDB
        try:
            self.template = ddb_util.get_prompt_data(context=self.context, aspect=aspect)        
            self.prompt = self.template.get('prompt', 'No prompt provided.')
        except Exception as e:
            print(e)
        
    def get_prompt(self, use_prompt=False):
        if use_prompt: return self.prompt

        template = """{guidelines}

Setting: {setting}

Persona: {persona}

Tone: {tone}

Task: {task}

Form: {form}

Examples: {examples}

Do not hallucinate - use only the context provided.

Answer: """

        prompt_template = PromptTemplate(
            input_variables=["guidelines", "setting", "persona", "task", "tone", "form", "examples"],
            template=template
        )

        formatted_prompt = prompt_template.format(
            guidelines=self.template.get('guidelines', 'No guidelines provided.'),
            setting=self.template.get('setting', 'No setting provided.'),
            persona=self.template.get('persona', 'No persona provided.'),
            task=self.template.get('task', 'No task provided.'),
            tone=self.template.get('tone', 'No tone provided.'),
            form=self.template.get('form', 'No form provided.'),
            examples='\n'.join(self.template.get('examples', ['No examples provided.']))
        )
        self.prompt = formatted_prompt
        return formatted_prompt

    def query(self, use_prompt=False, max_retries=3):
        retries = 0
        while retries < max_retries:
            prompt = self.get_prompt(use_prompt=use_prompt)
            answer = self.llm.query(prompt).answer
            if answer.lower() != 'i cannot answer.': return answer
            retries += 1
        return answer

    def post_doc(self, destination=None, use_prompt=False):
        fields = [{'table':self.context, 'row': self.aspect}]
        self.gu.post_doc_data(fields=fields, dest=destination, data=self.query(use_prompt=use_prompt)) 

    def post_sheet_query(self, sheet=None, column='D', q=query):
        fields = [{'table':self.context, 'row': self.aspect}]
        self.gu.post_sheet_data(fields=fields, sheet=sheet, column=column, data=query) 

    def post_sheet(self, sheet=None, column='D', use_prompt=False):
        fields = [{'table':self.context, 'row': self.aspect}]
        self.gu.post_sheet_data(fields=fields, sheet=sheet, column=column, data=self.query(use_prompt=use_prompt)) 

    def __str__(self):
        return json.dumps(self.template)
        
####################################################################

# fix URL
class DDBPromptWrapper(AspectFactory):
    def __init__(self, ddb_util=None, google_util=None, llm=None, context=None, aspect=None):
        super().__init__(ddb_util=ddb_util, google_util=google_util, llm=llm, context=context, aspect=aspect)

####################################################################

# Function to initialize the dynamic classes
def dynamic_init(self, context, aspect):
    super(self.__class__, self).__init__(context=context, aspect=aspect)   
    
# Function to get instances
def get_instances(aspects):
    # Create classes dynamically
    for aspect in aspects:
        for row in aspect['row']:
            class_name = aspect['table'] + row
            globals()[class_name] = type(class_name, (DDBP,), {"__init__": lambda self, context=aspect['table'], aspect=row: dynamic_init(self, context, aspect)})

    # Create instances dynamically
    return [globals()[aspect['table'] + row]() for aspect in aspects for row in aspect['row']]  




"""autopop: Auto Populate GDoc tables using LLMs."""

__version__ = '2.1'

from .aspect_factory import AspectFactory, DDBPromptWrapper, dynamic_init, get_instances
from .google_util_wrapper import GoogleUtilWrapper, call_post_sheet
from .ddb_util_wrapper import DDBUtilWrapper
from .knowledge_base import KnowledgeBase

__all__ = [
    'AspectFactory',
    'DDBPromptWrapper',
    'GoogleUtilWrapper',
    'DDBUtilWrapper',
    'KnowledgeBase',
    'call_post_sheet',
    'dynamic_init', 
    'get_instances'
]



import time
import requests

def call_post_sheet(instance, sheet=None, column=None, use_prompt=False): 
    instance.post_sheet(sheet=sheet, column=column, use_prompt=use_prompt)    

class GoogleUtilWrapper:
    def __init__(self, p2_url=None, gdt_url=None, pdest=None, gdest=None):
        self.p2_url = p2_url
        self.gdt_url = gdt_url
        self.pdest = pdest
        self.gdest = gdest

    def get_data(self, ep='', fields=None):
        payload = {}
        payload["p2"] = self.p2_url
        if fields: payload["fields"] = fields
        response = requests.get(self.gdt_url + ep, json=payload)
        return response.json()

    def post_doc_data(self, fields=[], data=None):
        payload = {}
        payload["p2"] = self.p2_url
        payload["fields"] = fields
        payload["destination"] = self.pdest
        if data: payload["data"] = data
        response = requests.post(self.gdt_url, json=payload)

    def post_sheet_data(self, fields=[], sheet=None, column=None, data=None):
        payload = {}
        payload["p2"] = self.p2_url
        payload["fields"] = fields
        payload["destination"] = self.gdest
        if data: payload["data"] = data
        if sheet: payload["sheet"] = sheet
        if column: payload["source"] = column
        response = requests.post(self.gdt_url, json=payload)